import numpy as np
import matplotlib.pyplot as plt
import numba as nb
# Assuming 'scabbard' is a custom module you have for loading data
# If not, you can replace it with an appropriate data loading mechanism
import scabbard as scb
import taichi as ti
from taichi.ui import Window, Canvas, GUI

###################
#### CONSTANTS ####
###################

## World up vector (Z-axis), constant
WORLD_UP = np.array([0, 0, 1])
## Background color in RGB 0-1
BGCOL_r = 0.
BGCOL_g = 0.
BGCOL_b = 0.

## shadow color in RGB 0-1
SHADOW_BLEND = 0.

## Render modes
'''
RENDER_COL is the mode of render coloration:
0: grayscale linked to normals (classic 3D mesh)
1: Water related
'''
RENDER_COL = 0

## Precision
RAY_STEP = 0.01      # Step size (smaller for higher accuracy)

## thresholds for plotting water
MIN_WATER = 0.01
## Therhold weight for water: minimum water height starts at MINWACOL * water col
MINWACOL = 0.5


@ti.func
def generate_realistic_water_colors():
	# Randomly generate blue and green dominant colors
	r = ti.random(ti.f32) * 50./255               # Low red to keep the color cool
	g = ti.random(ti.f32) * (80./255 )+ 100./255   # Mid-range green for teal/aqua tones
	b = ti.random(ti.f32) * (50./255) + 155./255  # High blue for a watery effect
	
	return ti.math.vec3(r,g,b)

@ti.func
def generic_water_colors():
	# Randomly generate blue and green dominant colors
	r = 30./255   # Low red to keep the color cool
	g = 80./255   # Mid-range green for teal/aqua tones
	b = 230./255  # High blue for a watery effect
	
	return ti.math.vec3(r,g,b)

@ti.func
def mix_rgb(color1:ti.math.vec3, color2:ti.math.vec3, weight:ti.f32 = 0.5):
	# Mix the two colors by weighted averaging each channel
	return color1 * weight + color2 * (1 - weight)


@ti.kernel
def reinhard_tone_mapping(image: ti.template(), PX: ti.template(), gamma: ti.f32):
	
	inv_gamma = 1.0 / gamma
	
	for i, j in PX:
		# Fetch the original color
		color = ti.Vector([image[i, j, 0], image[i, j, 1], image[i, j, 2]])
		R = color[0]
		G = color[1]
		B = color[2]

		# Apply Reinhard tone mapping to each channel
		R_mapped = R / (1.0 + R)
		G_mapped = G / (1.0 + G)
		B_mapped = B / (1.0 + B)

		# Apply gamma correction
		R_mapped = ti.pow(R_mapped, inv_gamma)
		G_mapped = ti.pow(G_mapped, inv_gamma)
		B_mapped = ti.pow(B_mapped, inv_gamma)

		# Write the mapped color to the output image
		image[i, j, 0] = R_mapped
		image[i, j, 1] = G_mapped
		image[i, j, 2] = B_mapped

@ti.func
def aces_tone_mapping(color: ti.types.vector(3, ti.f32)):
	# Constants for ACES approximation
	a = 2.51
	b = 0.03
	c = 2.43
	d = 0.59
	e = 0.14
	color_mapped = (color * (a * color + b)) / (color * (c * color + d) + e)
	return ti.math.clamp(color_mapped, 0.0, 1.0)

@ti.kernel
def aces_tone_mapping_kernel(image: ti.template(), PX:ti.template(), gamma: ti.f32):
	inv_gamma = 1.0 / gamma
	for i, j in PX:
		color = ti.Vector([image[i, j, 0], image[i, j, 1], image[i, j, 2]])
		# Apply ACES tone mapping
		color_mapped = aces_tone_mapping(color)
		# Apply gamma correction
		color_mapped = ti.pow(color_mapped, inv_gamma)
		image[i, j, 0] = color_mapped[0]
		image[i, j, 1] = color_mapped[1]
		image[i, j, 2] = color_mapped[2]

@ti.kernel
def toon_shader(image: ti.template(), PX:ti.template(), N_bin: ti.u8):
	for i, j in PX:
		color = ti.Vector([image[i, j, 0], image[i, j, 1], image[i, j, 2]])
		color *= N_bin
		color[0] = ti.math.round(color[0])
		color[1] = ti.math.round(color[1])
		color[2] = ti.math.round(color[2])

		image[i, j, 0] = color[0]/N_bin
		image[i, j, 1] = color[1]/N_bin
		image[i, j, 2] = color[2]/N_bin



@ti.func
def xy_to_ij(x:ti.template(), y:ti.template(), XX:ti.template(), YY:ti.template(), dx:ti.template(), dy:ti.template(), nx:ti.template(), ny:ti.template()):
	# Compute indices in grid
	j = (x - XX[0,0]) / dx
	i = (y - YY[0,0]) / dy

	i0 = int(ti.math.floor(i))
	j0 = int(ti.math.floor(j))

	# Ensure indices are within bounds
	i0 = max(0, min(i0, ny - 2))
	j0 = max(0, min(j0, nx - 2))

	return i, j, i0, j0



# Function to compute ray direction from camera through pixel
@ti.func
def compute_ray_direction(
	px:ti.template(), 
	py:ti.template(), 
	camera_position:ti.template(), 
	camera_direction:ti.template(), 
	camera_up:ti.template(), 
	camera_right:ti.template(), 
	focal_length:ti.f32
	):
	# Compute the point on the image plane in world coordinates
	# image_plane_point = ti.math.vec3(0.)
	image_plane_point = (camera_position +
						 camera_direction * focal_length +
						 camera_right * px +
						 camera_up * py)
	
	# Compute the ray direction from camera position to image plane point
	# ray_direction = ti.math.vec3(image_plane_point - camera_position[0], image_plane_point - camera_position[1], image_plane_point - camera_position[2])
	ray_direction = image_plane_point - camera_position
	norm = ti.math.length(ray_direction)
	ray_direction /= norm
	return ray_direction


@ti.func
def is_in_shadow(
	point: ti.math.vec3,
	light_direction: ti.math.vec3,
	XX: ti.template(),
	YY: ti.template(),
	ZZ: ti.template(),
) -> ti.u1:
	# Offset the point slightly to prevent self-intersection (shadow acne)
	bias = 1e-4
	shadow_origin = point + bias * light_direction
	# Cast a ray towards the light source
	t = 0.0
	max_t = 10.0  # Adjust based on scene size
	nx, ny = XX.shape[1], XX.shape[0]

	ret = 0
	while t < max_t:
		current_point = shadow_origin + t * light_direction
		x, y, z = current_point

		# Check if the point is within the scene bounds
		if x < XX[0, 0] or x > XX[0, nx - 1] or y < YY[0, 0] or y > YY[ny - 1, 0]:
			break  # Light is not obstructed within the scene

		# Get surface Z at (x, y)
		surface_z = interpolate_AA(x, y, XX, YY, ZZ)
		if surface_z != -9999.0 and z <= surface_z:
			ret = 1  # Point is in shadow
			break

		t += RAY_STEP

	return ret  # Point is not in shadow


# Function to interpolate ZZ at (x, y) using bilinear interpolation
@ti.func
def interpolate_AA(
	x:ti.template(), 
	y:ti.template(), 
	XX:ti.template(), 
	YY:ti.template(), 
	ZZ:ti.template()
	):
	
	nx,ny = XX.shape[1],XX.shape[0]
	# Ensure x and y are within the bounds
	z = -9999.
	if x < XX[0, 0] or x > XX[0, nx-1] or y < YY[0, 0] or y > YY[ny-1, 0]:
		z =  -9999.0
	else:
		# Find the spacing between grid points
		dx = XX[0,1] - XX[0,0]
		dy = YY[1,0] - YY[0,0]

		i,j,i0,j0 = xy_to_ij(x, y, XX, YY, dx, dy, nx, ny)

		i1 = i0 + 1
		j1 = j0 + 1

		# Compute fractional parts
		s = i - i0
		t = j - j0

		# Get Z values at the corners
		z00 = ZZ[i0, j0]
		z10 = ZZ[i1, j0]
		z01 = ZZ[i0, j1]
		z11 = ZZ[i1, j1]

		if(z00 == -9999.0 or z01 == -9999.0 or z10 == -9999.0 or z11 == -9999.0):
			z = -9999.0
		else:
			# Perform bilinear interpolation
			z0 = z00 * (1 - s) + z10 * s
			z1 = z01 * (1 - s) + z11 * s
			z  = z0  * (1 - t) + z1  * t

	return z

# Function to compute normal at (x, y) using gradients
@ti.func
def compute_normal(
	x:ti.template(),
	y:ti.template(), 
	XX:ti.template(), 
	YY:ti.template(), 
	ZZ:ti.template()
	):
	# Compute gradients using central differences
	h = 1e-5  # Small step for numerical derivative

	# Compute partial derivatives
	dz_dx = (interpolate_AA(x + h, y, XX, YY, ZZ) - interpolate_AA(x - h, y, XX, YY, ZZ)) / (2 * h)
	dz_dy = (interpolate_AA(x, y + h, XX, YY, ZZ) - interpolate_AA(x, y - h, XX, YY, ZZ)) / (2 * h)

	# The normal vector is [-dz/dx, -dz/dy, 1]
	normal = ti.math.vec3(-dz_dx, -dz_dy, 1.0)
	norm = ti.math.length(normal)
	normal /= norm

	return normal

# Function to perform ray-surface intersection using ray marching
@ti.func
def ray_surface_intersection(
	ray_origin:ti.math.vec3, 
	ray_direction:ti.math.vec3,
	XX:ti.template(),
	YY:ti.template(),
	ZZ:ti.template(),

	):
	# Initialize t (ray parameter)
	t = 0.0   # Start from t = 0.0
	max_t = 10.0  # Maximum distance to march
	prev_dz = -9999.0
	intersection_point = ti.math.vec3(-9999.)
	normal = ti.math.vec3(-9999.)
	dz = 10.
	nx,ny = XX.shape[1],XX.shape[0]
	cont = True
	xf = -9999.
	while t < max_t and cont:
		# Compute current point along the ray
		point = ray_origin + t * ray_direction
		x, y, z = point

		# Get surface Z at (x, y) if within domain
		if x >= XX[0,0] and x <= XX[0,nx-1] and y >= YY[0,0] and y <= YY[ny-1,0]:
			surface_z = interpolate_AA(x, y, XX, YY, ZZ)
			if surface_z != -9999.0 and z != -9999.0:
				# Compute difference between ray's z and surface z
				dz = z - surface_z

				if prev_dz != -9999.0:
					if dz * prev_dz < 0:
						# The sign of dz changed, indicating a crossing
						# Perform linear interpolation to find intersection t
						t_intersect = t - RAY_STEP * dz / (dz - prev_dz)
						# Compute intersection point
						intersection_point = ray_origin + t_intersect * ray_direction
						x_int, y_int, z_int = intersection_point
						# Compute normal at intersection point
						normal = compute_normal((x_int), (y_int), XX, YY, ZZ)
						cont = False
					elif dz == 0.0:
						# Ray is exactly on the surface
						intersection_point = point
						normal = compute_normal(x, y,XX,YY,ZZ)
						cont = False

				prev_dz = dz
			else:
				# Surface Z is invalid
				pass
		else:
			# If the ray is outside the domain and moving away, terminate early
			if prev_dz != -9999.0 and dz > 0:
				break  # Ray is above the surface and moving upwards

		t += RAY_STEP

	# No intersection found
	return intersection_point, normal

@ti.func
def compute_ray_ortho(
	px: ti.template(),
	py: ti.template(),
	camera_position: ti.template(),
	camera_direction: ti.template(),
	camera_up: ti.template(),
	camera_right: ti.template(),
):
	# Ray direction is constant in orthographic projection
	ray_direction = camera_direction

	# Ray origin varies across the image plane
	ray_origin = (camera_position +
				  camera_right * px +
				  camera_up * py)
	return ray_origin, ray_direction

@ti.func
def compute_ray_ortho(
	px: ti.template(),
	py: ti.template(),
	camera_position: ti.template(),
	camera_direction: ti.template(),
	camera_up: ti.template(),
	camera_right: ti.template(),
	dx:ti.f32,
	dy:ti.f32
):
	# Ray direction is constant in orthographic projection
	ray_direction = ti.math.vec3(camera_direction)

	# Ray origin varies across the image plane
	ray_origin = ti.math.vec3(camera_position +
				  camera_right * (px + dx) +
				  camera_up * (py + dy))
	return ray_origin, ray_direction

@ti.func
def compute_ray_persp(
	px: ti.template(),
	py: ti.template(),
	camera_position: ti.template(),
	camera_direction: ti.template(),
	camera_up: ti.template(),
	camera_right: ti.template(),
	focal_length:ti.f32,
	dx:ti.f32,
	dy:ti.f32
):
	ray_origin = ti.math.vec3(camera_position[0],camera_position[1],camera_position[2])
	# Compute ray direction
	ray_direction = compute_ray_direction(px+dx, py+dy, camera_position,camera_direction,camera_up,camera_right,focal_length)
	return ray_origin, ray_direction

	

@ti.kernel
def render_gpu( 
	XX:ti.template(),
	YY:ti.template(),
	ZZ:ti.template(),
	HH:ti.template(),
	image:ti.template(),
	PX:ti.template(),
	PY:ti.template(),
	camera_position:ti.template(),
	camera_direction:ti.template(),
	camera_up:ti.template(),
	camera_right:ti.template(),
	focal_length:ti.f32, 
	image_height:ti.i32, 
	image_plane_height:ti.f32, 
	image_width:ti.i32,
	image_plane_width:ti.f32,
	N_AA:ti.i32,
	ortho:ti.u1,

	):
	
	pixel_width = (image_plane_width / image_width)
	pixel_height = (image_plane_height / image_height)
	nx,ny = XX.shape[1],XX.shape[0]

	# Simple shading using Lambertian reflection
	# Define light direction (e.g., from above)
	light_direction = ti.math.vec3(1.0, 1.0, 0.75)  # Light coming from (1,1,1)
	light_direction /= ti.math.length(light_direction)

	# Loop over each pixel in the image
	for i,j in ti.ndrange((0,image_height),(0,image_width)):
		# Compute pixel coordinates in image plane
		px = PX[i, j]
		py = PY[i, j]
		
		color = ti.math.vec3(0.)
		nav = 0
		ray_origin = ti.math.vec3(0.) 
		ray_direction = ti.math.vec3(0.)

		save_intersection_point = ti.math.vec3(0.)
		has_intersection = False

		for kk in range(N_AA):

			# Generate random offsets within the pixel
			dx = (ti.random(ti.f32) - 0.5) * pixel_width
			dy = (ti.random(ti.f32) - 0.5) * pixel_height


			# Ray origin is the camera position
			if(ortho == False):
				ray_origin, ray_direction = compute_ray_persp(px,py,camera_position,camera_direction,camera_up,camera_right,focal_length,dx,dy)	
			else:
				ray_origin, ray_direction = compute_ray_ortho(px,py,camera_position,camera_direction,camera_up,camera_right,dx,dy)	

			# Perform ray-surface intersection
			intersection_point, normal = ray_surface_intersection(ray_origin, ray_direction, XX, YY, ZZ)
			
			if intersection_point[0] != -9999.0:
				has_intersection = True
				save_intersection_point = intersection_point

				# Compute intensity
				intensity = ti.math.dot(normal, light_direction)
				intensity = max(0.01, min(intensity, 1.0))

				tcolor = ti.math.vec3(0.)
				# Assign color based on intensity
				if RENDER_COL == 0:
					tcolor += intensity * ti.math.vec3(1.0, 1.0, 1.0)  # White color scaled by intensity

				elif RENDER_COL == 1:
					x,y = intersection_point[0], intersection_point[1]
					tti, ttj, i0, j0 = xy_to_ij(x, y, XX, YY, dx, dy, nx, ny)
					tcolor += intensity * ti.math.vec3(1.0, 1.0, 1.0)
					th = interpolate_AA(x,y, XX, YY, HH)
					
					if(th > MIN_WATER):
						weight = max(MINWACOL, min(1., th/2.))
						tcolor = mix_rgb(tcolor, generic_water_colors(), (1. - weight))

				
				color += tcolor
				nav += 1

		color /= ti.max(1,nav)

		if(has_intersection and  SHADOW_BLEND > 0):
			if(is_in_shadow(save_intersection_point, light_direction, XX, YY, ZZ)):
				color = mix_rgb(color, ti.math.vec3(0.), SHADOW_BLEND)

		if(color[0] > 0 or color[1] > 0 or color[2] > 0):
			image[i, j, 0] = color[0]
			image[i, j, 1] = color[1]
			image[i, j, 2] = color[2]
		else:
			image[i, j, 0] = BGCOL_r
			image[i, j, 1] = BGCOL_g
			image[i, j, 2] = BGCOL_b





# ==========================================
# Step 6: Main Rendering Loop
# ==========================================


@ti.kernel
def rotate(
	tp:ti.template(),
	image:ti.template(),
	PX:ti.template()
	):

	for i,j in PX:
		tp[j,i,0] = image[i,j,0] 
		tp[j,i,1] = image[i,j,1] 
		tp[j,i,2] = image[i,j,2]







class RT_data:

	def __init__(self, grid, exaggeration_factor = 0.5, hwater = None):


		# ==========================================
		# Step 1: Data Preparation
		# ==========================================
		self.grid = grid

		Z = self.grid.Z[1:-1,1:-1]

		self._exaggeration_factor = exaggeration_factor

		mask = Z!=-9999

		# Normalize Z to range from 0 to 1, then apply exaggeration factor
		Z_min, Z_max = Z[mask].min(), Z[mask].max()
		Z_normalized = np.copy(Z)
		Z_normalized[mask] = (Z[mask] - Z_min) / (Z_max - Z_min) * self._exaggeration_factor


		self.ny, self.nx = Z.shape
		self.x = np.linspace(-1, 1, self.nx)
		rat = self.ny/self.nx
		self.y = np.linspace(-1*rat, 1*rat, self.ny)
		_XX, _YY = np.meshgrid(self.x, self.y)

		# Set ZZ to the normalized Z values
		_ZZ = Z_normalized

		# Compute the center point (0, 0, median ZZ)
		self.center_Z = np.median(_ZZ[mask])
		self.center_point = np.array([0.0, 0.0, self.center_Z])

		self.XX = ti.field(ti.f32, shape = _XX.shape)
		self.XX.from_numpy(_XX.astype(np.float32))
		self.YY = ti.field(ti.f32, shape = _YY.shape)
		self.YY.from_numpy(_YY.astype(np.float32))
		self.ZZ = ti.field(ti.f32, shape = _ZZ.shape)

		if(hwater is not None):
			th = (np.copy(hwater[1:-1,1:-1])) / (Z_max - Z_min) * self._exaggeration_factor
			self.HH = ti.field(ti.f32, shape = _ZZ.shape)
			self.HH.from_numpy(hwater[1:-1,1:-1].astype(np.float32))
			self.ZZ.from_numpy(_ZZ.astype(np.float32) + th.astype(np.float32))
		else:
			# Dummy array. Needed.
			self.HH = ti.field(ti.f32, shape = (4,4))
			self.HH.fill(0.)
			self.ZZ.from_numpy(_ZZ.astype(np.float32))



class RT_camera:
	def __init__(self,
		rtdata,
		camera_distance = 2.0,      # Distance from the center (adjust for zoom)
		camera_azimuth_deg = 270,   # Azimuth angle in degrees
		camera_elevation_deg = 45,  # Elevation angle from horizontal in degrees):
	):
		'''
		Sets a first time the camera attributes
		'''
		self.rtdata = rtdata
		self.camera_distance = None
		self.camera_azimuth_deg = None
		self.camera_elevation_deg = None
		self.camera_position = ti.Vector(np.zeros(3, dtype = np.float32))
		self.camera_direction = ti.Vector(np.zeros(3, dtype = np.float32))
		self.camera_right = ti.Vector(np.zeros(3, dtype = np.float32))
		self.camera_up  = ti.Vector(np.zeros(3, dtype = np.float32))

		self.update(camera_distance,camera_azimuth_deg,camera_elevation_deg)


		

	def update(self,
		camera_distance,
		camera_azimuth_deg,
		camera_elevation_deg,
	):

		self.camera_distance = camera_distance
		self.camera_azimuth_deg = camera_azimuth_deg
		self.camera_elevation_deg = camera_elevation_deg

		# Convert angles to radians
		theta = np.deg2rad(self.camera_azimuth_deg)      # Azimuth angle in radians
		phi = np.deg2rad(self.camera_elevation_deg)      # Elevation angle in radians

		# Compute self.camera position using spherical coordinates
		camera_x = self.rtdata.center_point[0] + self.camera_distance * np.cos(phi) * np.cos(theta)
		camera_y = self.rtdata.center_point[1] + self.camera_distance * np.cos(phi) * np.sin(theta)
		camera_z = self.rtdata.center_point[2] + self.camera_distance * np.sin(phi)
		_camera_position = np.array([camera_x, camera_y, camera_z])

		# Camera direction vector (from camera position to center point)
		_camera_direction = self.rtdata.center_point - _camera_position
		_camera_direction /= np.linalg.norm(_camera_direction)  # Normalize

		# Camera right vector
		_camera_right = np.cross(_camera_direction, WORLD_UP)
		_camera_right /= np.linalg.norm(_camera_right)  # Normalize

		# Camera up vector
		_camera_up = np.cross(_camera_right, _camera_direction)
		_camera_up /= np.linalg.norm(_camera_up)

		_camera_position = np.array([camera_x, camera_y, camera_z])

		self.camera_position = ti.Vector(_camera_position.astype(np.float32))
		self.camera_direction = ti.Vector(_camera_direction.astype(np.float32))
		self.camera_right = ti.Vector(_camera_right.astype(np.float32))
		self.camera_up = ti.Vector(_camera_up.astype(np.float32))


class RT_image:

	def __init__(self,
		image_width = 1200, 
		image_height = 900,
		focal_length = 1.0,         # Adjust as needed
		fov_deg = 60,               # Field of view in degrees (reduced to minimize distortion)
		
		ortho = False,
		ortho_height = 5.
		):

		self.image_width = image_width
		self.image_height = image_height
		self.focal_length = focal_length
		self.fov_deg = fov_deg
		# Set up the image plane dimensions based on field of view and aspect ratio
		self.aspect_ratio = image_width / image_height
		self.fov_rad = np.deg2rad(fov_deg)  # Convert to radians


		self.ortho = ortho
		self.ortho_height = ortho_height
		
		# Compute image plane dimensions
		if(self.ortho):
			self.image_plane_height = self.ortho_height
			self.image_plane_width = self.image_plane_height * self.aspect_ratio
		else:
			self.image_plane_height = 2 * self.focal_length * np.tan(self.fov_rad / 2)
			self.image_plane_width = self.image_plane_height * self.aspect_ratio

		# Generate pixel coordinates in the image plane
		self.px = np.linspace(-0.5 * self.image_plane_width, 0.5 * self.image_plane_width, self.image_width)
		self.py = np.linspace(-0.5 * self.image_plane_height, 0.5 * self.image_plane_height, self.image_height)
		_PX, _PY = np.meshgrid(self.px, self.py)

		self.PX = ti.field(ti.f32, shape = _PX.shape)
		self.PX.from_numpy(_PX.astype(np.float32))
		self.PY = ti.field(ti.f32, shape = _PY.shape)
		self.PY.from_numpy(_PY.astype(np.float32))
		
		self.image = ti.field(ti.f32, shape = (image_height, image_width, 3) )
		self.image.fill(0.)

class RT_renderer:

	def __init__(self,
		N_AA = 5
	):
		self.N_AA = N_AA

def _render_RT(
	rtdata,
	camera,
	image,
	renderer,
	which = 'gray'
):
	'''
	Internal wrapper for the gray renderer operating from the RT objects
	'''
	render_gpu(
		rtdata.XX,rtdata.YY,rtdata.ZZ,rtdata.HH,
		image.image,image.PX,image.PY,
		camera.camera_position,camera.camera_direction,camera.camera_up,camera.camera_right,
		image.focal_length,image.image_height,image.image_plane_height,image.image_width,image.image_plane_width, 
		renderer.N_AA, image.ortho
	)



def std_gray_RT(
	grid:scb.raster.RegularRasterGrid,
	exaggeration_factor = 0.5,  # Adjust this as needed
	camera_distance = 2.0,      # Distance from the center (adjust for zoom)
	camera_azimuth_deg = 270,   # Azimuth angle in degrees
	camera_elevation_deg = 45,  # Elevation angle from horizontal in degrees
	focal_length = 1.0,         # Adjust as needed
	image_width = 1200,         # Image width in pixels
	image_height = 900,         # Image height in pixels
	fov_deg = 60,               # Field of view in degrees (reduced to minimize distortion)
	N_AA = 5,
	tone_mapping = False,
	toon = 0,
	ortho = False,
	ortho_height = 5.,
	hw = None

	):

	ti.init(ti.gpu)

	rtdata = RT_data(grid, exaggeration_factor = exaggeration_factor) if (hw is None) else RT_data(grid, exaggeration_factor = exaggeration_factor, hwater = hw)

	camera = RT_camera(rtdata, camera_distance = camera_distance, camera_azimuth_deg = camera_azimuth_deg, camera_elevation_deg = camera_elevation_deg)
	image = RT_image(image_width = image_width, image_height = image_height, focal_length = focal_length, fov_deg = fov_deg, ortho = ortho,ortho_height = ortho_height)
	renderer = RT_renderer(N_AA = N_AA)

	_render_RT(rtdata, camera, image, renderer)

	if(tone_mapping):
		# reinhard_tone_mapping(image.image, image.PX, 2.2)
		aces_tone_mapping_kernel(image.image, image.PX, 2.2)

	if(toon > 0):
		toon_shader(image.image, image.PX, toon)

	return image.image.to_numpy()[::-1]

def std_water_RT(
	grid:scb.raster.RegularRasterGrid,
	hw:np.ndarray,
	exaggeration_factor = 0.5,  # Adjust this as needed
	camera_distance = 2.0,      # Distance from the center (adjust for zoom)
	camera_azimuth_deg = 270,   # Azimuth angle in degrees
	camera_elevation_deg = 45,  # Elevation angle from horizontal in degrees
	focal_length = 1.0,         # Adjust as needed
	image_width = 1200,         # Image width in pixels
	image_height = 900,         # Image height in pixels
	fov_deg = 60,               # Field of view in degrees (reduced to minimize distortion)
	N_AA = 5,
	tone_mapping = False,
	toon = 0

	):

	ti.init(ti.gpu)

	rtdata = RT_data(grid, exaggeration_factor = exaggeration_factor, hwater = hw)
	camera = RT_camera(rtdata, camera_distance = camera_distance, camera_azimuth_deg = camera_azimuth_deg, camera_elevation_deg = camera_elevation_deg)
	image = RT_image(image_width = image_width, image_height = image_height, focal_length = focal_length, fov_deg = fov_deg )
	renderer = RT_renderer(N_AA = N_AA)

	_render_RT(rtdata, camera, image, renderer, which = 'water1')

	if(tone_mapping):
		# reinhard_tone_mapping(image.image, image.PX, 2.2)
		aces_tone_mapping_kernel(image.image, image.PX, 2.2)

	if(toon > 0):
		toon_shader(image.image, image.PX, toon)

	return image.image.to_numpy()[::-1]






def set_bg_constants(r,g,b):
	'''
	To be called before any RT operation
	Sets the constants for the background color in the RT visu
	'''
	global BGCOL_r, BGCOL_g, BGCOL_b
	BGCOL_r = r
	BGCOL_g = g
	BGCOL_b = b






def set_render_mode(which = 'gray'):
	global RENDER_COL
	if(which.lower() == 'gray'):
		RENDER_COL = 0
	elif(which.lower() == 'water'):
		RENDER_COL = 1






















#########################################################
#########################################################
#########################################################
################# LEGACY ################################
#########################################################
#########################################################
#########################################################





























def yolo(
	grid:scb.raster.RegularRasterGrid,
	exaggeration_factor = 0.5,  # Adjust this as needed
	camera_distance = 2.0,      # Distance from the center (adjust for zoom)
	camera_azimuth_deg = 270,   # Azimuth angle in degrees
	camera_elevation_deg = 45,  # Elevation angle from horizontal in degrees
	focal_length = 1.0,         # Adjust as needed
	image_width = 1200,         # Image width in pixels
	image_height = 900,         # Image height in pixels
	fov_deg = 60,               # Field of view in degrees (reduced to minimize distortion)
	N_AA = 10
):

	
	ti.init(ti.gpu)

	rtdata = RT_data(grid, exaggeration_factor = exaggeration_factor)
	camera = RT_camera(rtdata, camera_distance = camera_distance, camera_azimuth_deg = camera_azimuth_deg, camera_elevation_deg = camera_elevation_deg)
	image = RT_image(image_width = image_width, image_height = image_height, focal_length = focal_length, fov_deg = fov_deg )
	renderer = RT_renderer(N_AA = N_AA)

	window = ti.ui.Window("YOLO", (image_width, image_height))
	canvas = window.get_canvas()
	gui = window.get_gui()

	out = ti.field(ti.f32, shape = (image_height, image_width, 3) )

	while window.running:


		# Start GUI
		gui.begin("Settings", 0.05, 0.05, 0.3, 0.3)
		gui.text("Camera Controls")

		camera_azimuth_deg = gui.slider_float("Azimuth", camera_azimuth_deg, 0.0, 360.0)
		camera_elevation_deg = gui.slider_float("Elevation", camera_elevation_deg, 0.0, 90.0)
		camera_distance = gui.slider_float("Distance", camera_distance, 0.1, 10.0)
		
		gui.end()

		# Update camera parameters
		camera.update(camera_distance, camera_azimuth_deg, camera_elevation_deg)

		# Render the image
		_gray_RT(rtdata, camera, image, renderer)
		rotate(out, image.image, image.PX)
		# Draw the image onto the canvas
		canvas.set_image(out)

		# Show the frame
		window.show()




# LEGACY RT, to keep while the others are not ready
def legacy_gray_RT(
	grid:scb.raster.RegularRasterGrid,
	exaggeration_factor = 0.5,  # Adjust this as needed
	camera_distance = 2.0,      # Distance from the center (adjust for zoom)
	camera_azimuth_deg = 270,   # Azimuth angle in degrees
	camera_elevation_deg = 45,  # Elevation angle from horizontal in degrees
	focal_length = 1.0,         # Adjust as needed
	image_width = 1200,         # Image width in pixels
	image_height = 900,         # Image height in pixels
	fov_deg = 60,               # Field of view in degrees (reduced to minimize distortion)
	N_AA = 5,

	):

	ti.init(ti.gpu)

	# ==========================================
	# Step 1: Data Preparation
	# ==========================================

	Z = grid.Z[1:-1,1:-1]

	# Normalize Z to range from 0 to 1, then apply exaggeration factor
	Z_min, Z_max = Z.min(), Z.max()
	Z_normalized = (Z - Z_min) / (Z_max - Z_min) * exaggeration_factor

	# Generate XX and YY coordinates ranging from -1 to 1
	N, M = Z.shape
	ny, nx = Z.shape
	x = np.linspace(-1, 1, M)
	y = np.linspace(-1, 1, N)
	_XX, _YY = np.meshgrid(x, y)

	# Set ZZ to the normalized Z values
	_ZZ = Z_normalized

	# ==========================================
	# Step 2: Camera Setup
	# ==========================================

	# Compute the center point (0, 0, median ZZ)
	center_Z = np.median(_ZZ)
	center_point = np.array([0.0, 0.0, center_Z])



	# Convert angles to radians
	theta = np.deg2rad(camera_azimuth_deg)      # Azimuth angle in radians
	phi = np.deg2rad(camera_elevation_deg)      # Elevation angle in radians

	# Compute camera position using spherical coordinates
	camera_x = center_point[0] + camera_distance * np.cos(phi) * np.cos(theta)
	camera_y = center_point[1] + camera_distance * np.cos(phi) * np.sin(theta)
	camera_z = center_point[2] + camera_distance * np.sin(phi)
	_camera_position = np.array([camera_x, camera_y, camera_z])

	# ==========================================
	# Step 3: Camera Coordinate System
	# ==========================================

	# Camera direction vector (from camera position to center point)
	_camera_direction = center_point - _camera_position
	_camera_direction /= np.linalg.norm(_camera_direction)  # Normalize

	# World up vector (Z-axis)
	WORLD_UP = np.array([0, 0, 1])

	# Camera right vector
	_camera_right = np.cross(_camera_direction, WORLD_UP)
	_camera_right /= np.linalg.norm(_camera_right)  # Normalize

	# Camera up vector
	_camera_up = np.cross(_camera_right, _camera_direction)
	_camera_up /= np.linalg.norm(_camera_up)

	# ==========================================
	# Step 4: Image Plane Setup
	# ==========================================

	

	# Set up the image plane dimensions based on field of view and aspect ratio
	aspect_ratio = image_width / image_height
	fov_rad = np.deg2rad(fov_deg)  # Convert to radians

	# Compute image plane dimensions
	image_plane_height = 2 * focal_length * np.tan(fov_rad / 2)
	image_plane_width = image_plane_height * aspect_ratio

	# Generate pixel coordinates in the image plane
	px = np.linspace(-0.5 * image_plane_width, 0.5 * image_plane_width, image_width)
	py = np.linspace(-0.5 * image_plane_height, 0.5 * image_plane_height, image_height)
	_PX, _PY = np.meshgrid(px, py)



	PX = ti.field(ti.f32, shape = _PX.shape)
	PX.from_numpy(_PX.astype(np.float32))
	PY = ti.field(ti.f32, shape = _PY.shape)
	PY.from_numpy(_PY.astype(np.float32))
	
	image = ti.field(ti.f32, shape = (image_height, image_width, 3) )
	image.fill(0.)
	
	out = ti.field(ti.f32, shape = (image_height, image_width, 3) )

	XX = ti.field(ti.f32, shape = _XX.shape)
	XX.from_numpy(_XX.astype(np.float32))
	YY = ti.field(ti.f32, shape = _YY.shape)
	YY.from_numpy(_YY.astype(np.float32))
	ZZ = ti.field(ti.f32, shape = _ZZ.shape)
	ZZ.from_numpy(_ZZ.astype(np.float32))

	camera_position = ti.Vector(_camera_position.astype(np.float32))
	camera_direction = ti.Vector(_camera_direction.astype(np.float32))
	camera_right = ti.Vector(_camera_right.astype(np.float32))
	camera_up = ti.Vector(_camera_up.astype(np.float32))


	render_gray(XX,YY,ZZ,image,PX,PY,camera_position,camera_direction,camera_up,camera_right,focal_length,image_height,image_plane_height,image_width,image_plane_width, N_AA)
	
	# rotate(out, image, PX)


	return image.to_numpy()[::-1]




@ti.kernel
def render_gray( 
	XX:ti.template(),
	YY:ti.template(),
	ZZ:ti.template(),
	image:ti.template(),
	PX:ti.template(),
	PY:ti.template(),
	camera_position:ti.template(),
	camera_direction:ti.template(),
	camera_up:ti.template(),
	camera_right:ti.template(),
	focal_length:ti.f32, 
	image_height:ti.i32, 
	image_plane_height:ti.f32, 
	image_width:ti.i32,
	image_plane_width:ti.f32,
	N_AA:ti.i32,

	):
	
	pixel_width = (image_plane_width / image_width)
	pixel_height = (image_plane_height / image_height)
	nx,ny = XX.shape[1],XX.shape[0]

	# Loop over each pixel in the image
	for i,j in ti.ndrange((0,image_height),(0,image_width)):
		# Compute pixel coordinates in image plane
		px = PX[i, j]
		py = PY[i, j]
		

		# Ray origin is the camera position
		ray_origin = ti.math.vec3(camera_position[0],camera_position[1],camera_position[2])

		color = ti.math.vec3(0.)
		nav = 0

		for kk in range(N_AA):

			# Generate random offsets within the pixel
			dx = (ti.random(ti.f32) - 0.5) * pixel_width
			dy = (ti.random(ti.f32) - 0.5) * pixel_height

			# Compute ray direction
			ray_direction = compute_ray_direction(px+dx, py+dy, camera_position,camera_direction,camera_up,camera_right,focal_length)

			# Perform ray-surface intersection
			intersection_point, normal = ray_surface_intersection(ray_origin, ray_direction, XX, YY, ZZ)
			
			if intersection_point[0] != -9999.0:
				# Simple shading using Lambertian reflection
				# Define light direction (e.g., from above)
				light_direction = ti.math.vec3(1.0, 1.0, 1.0)  # Light coming from (1,1,1)
				light_direction /= ti.math.length(light_direction)

				# Compute intensity
				intensity = ti.math.dot(normal, light_direction)
				intensity = max(0.0, min(intensity, 1.0))

				# Assign color based on intensity
				color += intensity * ti.math.vec3(1.0, 1.0, 1.0)  # White color scaled by intensity
				nav += 1

		color /= ti.max(1,nav)

		image[i, j, 0] = color[0]
		image[i, j, 1] = color[1]
		image[i, j, 2] = color[2]

@ti.kernel
def render_with_water( 
	XX:ti.template(),
	YY:ti.template(),
	ZZ:ti.template(),
	HH:ti.template(),
	image:ti.template(),
	PX:ti.template(),
	PY:ti.template(),
	camera_position:ti.template(),
	camera_direction:ti.template(),
	camera_up:ti.template(),
	camera_right:ti.template(),
	focal_length:ti.f32, 
	image_height:ti.i32, 
	image_plane_height:ti.f32, 
	image_width:ti.i32,
	image_plane_width:ti.f32,
	N_AA:ti.i32,

	):
	
	nx,ny = XX.shape[1], XX.shape[0]
	XX[0,1] - XX[0,0], YY[0,1] - YY[0,0]
	pixel_width = (image_plane_width / image_width)
	pixel_height = (image_plane_height / image_height)

	# Loop over each pixel in the image
	for i,j in ti.ndrange((0,image_height),(0,image_width)):
		# Compute pixel coordinates in image plane
		px = PX[i, j]
		py = PY[i, j]
		

		# Ray origin is the camera position
		ray_origin = ti.math.vec3(camera_position[0],camera_position[1],camera_position[2])

		color = ti.math.vec3(0.)
		nav = 0

		for kk in range(N_AA):

			# Generate random offsets within the pixel
			dx = (ti.random(ti.f32) - 0.5) * pixel_width
			dy = (ti.random(ti.f32) - 0.5) * pixel_height

			# Compute ray direction
			ray_direction = compute_ray_direction(px+dx, py+dy, camera_position,camera_direction,camera_up,camera_right,focal_length)

			# Perform ray-surface intersection
			intersection_point, normal = ray_surface_intersection(ray_origin, ray_direction, XX, YY, ZZ)
			
			if intersection_point[2] != -9999.0:

				# Simple shading using Lambertian reflection
				# Define light direction (e.g., from above)
				light_direction = ti.math.vec3(1.0, 1.0, 1.0)  # Light coming from (1,1,1)
				light_direction /= ti.math.length(light_direction)

				# Compute intensity
				intensity = ti.math.dot(normal, light_direction)
				intensity = max(0.0, min(intensity, 1.0))
				x,y = intersection_point[0], intersection_point[1]

				tcolor = ti.math.vec3(0.,0.,0.)
				if x >= XX[0,0] and x <= XX[0,nx-1] and y >= YY[0,0] and y <= YY[ny-1,0]:
					tti, ttj, i0, j0 = xy_to_ij(x, y, XX, YY, dx, dy, nx, ny)
					# print(i0,j0)

					tcolor = intensity * ti.math.vec3(1.0, 1.0, 1.0)

					weight = 1. if HH[i0,j0] > 0.05 else 0.

					tz = interpolate_AA(x,y, XX, YY, ZZ)
					th = interpolate_AA(x,y, XX, YY, HH)
					# if(HH[i0,j0]> 1.):
					# 	print(i0,j0,HH[i0,j0])
					weight = min(1., th/2.)

					tcolor = mix_rgb(tcolor, generate_realistic_water_colors(), (1. - weight))
					# tcolor = generic_water_colors() if th > 0.05 else tcolor
					# tcolor = ti.math.vec3(1.0, 1.0, 1.0) * interpolate_AA(x,y,XX,YY,ZZ)

					# Assign color based on intensity
					color += tcolor # White color scaled by intensity
					nav += 1

		color /= ti.max(1,nav)
		image[i, j, 0] = color[0]
		image[i, j, 1] = color[1]
		image[i, j, 2] = color[2]


import scabbard as scb
import numpy as np
import numba as nb



class RivNet:
	'''
	Rivet is a simple river network base class holding the minimal data structure to define a river network
	'''

	def __init__(self):

		# Holds the node flat IDs from the original 
		self.nodes = None


'''
Help with managing boundary conditions
Weither it is about automatically generating them or manage more complex cases like no data and all


B.G.
'''

import numpy as np
import scabbard._utils as ut
import scabbard.flow._bc_helper as bch
from scabbard.flow.graph import SFGraph
import scabbard as scb
import dagger as dag
import numba as nb
from functools import reduce
from collections.abc import Iterable


def get_normal_BCs(dem):
	'''
	Returns an array of boundary conditions with "normal edges"
	Flow can out at each edge

	parameters:
		- can be a 2D numpy array or a raster object
	
	Authors:
	- B.G (alst modification: 07/2024)
	'''
	if(isinstance(dem,np.ndarray)):
		return ut.normal_BCs_from_shape(dem.shape[1], dem.shape[0])
	elif(isinstance(dem,scb.raster.RegularRasterGrid)):
		return ut.normal_BCs_from_shape(dem.geo.nx, dem.geo.ny)
	else:
		raise TypeError('boundary conditions can only be obtained from a RegularRasterGrid or a 2D numpy array')

def combine_masks(*args):
	return reduce(np.bitwise_and, args) if all(isinstance(arr, np.ndarray) and arr.dtype in [np.uint8, np.bool_] for arr in args) else None

@nb.njit()
def _mask_to_BCs(BCs, mask, nx, ny):
	'''
	Internal numba function helping with converting a mask of partial nodata to BCs codes

	Arguments:
		- BCs: the 2D array of BCs
		- mask: the binary mask  of reference
		- nx and ny: the number of cols and rows
	Returns:
		- Nothing, edits in place

	Author:
		- B.G. (last modification: 09/2024 for MTJ project)
	'''

	# First pass to mark nodata
	for i in range(ny):
		for j in range(nx):
			if mask[i,j] == 0:
				BCs[i,j] = 0

	# second to find outlet
	for i in range(ny):
		for j in range(nx):

			if(BCs[i,j] == 1):
				for k in range(4):
					ir,jr = scb.ste.neighbours_D4(i,j,k,BCs,nx,ny)
					if(ir == -1):
						BCs[i,j] = 3
				


def mask_to_BCs(grid, mask):
	'''
	Converts a mask of partial nodata to BCs codes

	Arguments:
		- grid: raster grid
		- mask: the binary mask of reference
	Returns:
		- The 2D array of boundary codes

	Author:
		- B.G. (last modification: 09/2024 for MTJ project)
	'''

	if mask is None:
		raise ValueError('scabbard.flow.bc_helper.mask_to_BCs::Mask is None')

	tgrid = grid.Z if isinstance(grid, scb.raster.RegularRasterGrid) else grid.Z2D
	ny,nx = (grid.geo.ny, grid.geo.nx) if isinstance(grid, scb.raster.RegularRasterGrid) else (grid.ny, grid.nx)
	dx = grid.geo.dx if isinstance(grid, scb.raster.RegularRasterGrid) else grid.dx

	# Preprocessing the boundary conditions
	gridcpp = dag.GridCPP_f32(nx,ny,dx,dx,3)
	BCs = np.ones_like(tgrid, dtype = np.uint8)
	# dag.mask_to_BCs_f32(gridcpp, mask, BCs, False)
	_mask_to_BCs(BCs,mask,nx,ny)

	BCs[[0,-1],:][BCs[[0,-1],:] > 0] = 3
	BCs[:, [0,-1]][BCs[:, [0,-1]] > 0] = 3

	return BCs


def mask_seas(grid, sea_level = 0., extra_mask = None):
	'''
	Returns or append a mask with 0s where  
	'''
	
	tgrid = grid.Z if isinstance(grid, scb.raster.RegularRasterGrid) else grid.Z2D

	mask = np.ones_like(tgrid, dtype = np.uint8)
	mask[tgrid < sea_level] = 0
	mask[np.isfinite(tgrid) == False] = 0
	
	if (extra_mask is None):
		return mask
	else:
		return (extra_mask & mask)

def mask_single_watershed_from_outlet(grid, location, BCs = None, extra_mask = None, MFD = True, stg = None):

	tgrid = grid.Z if isinstance(grid, scb.raster.RegularRasterGrid) else grid.Z2D
	ny,nx = (grid.geo.ny, grid.geo.nx) if isinstance(grid, scb.raster.RegularRasterGrid) else (grid.ny, grid.nx)
	dx = grid.geo.dx if isinstance(grid, scb.raster.RegularRasterGrid) else grid.dx

	# Checks if the input is flat index or rows col
	if(isinstance(location, Iterable) and not isinstance(location, (str, bytes))):
		row,col = location
		index = row * nx + col
	else:
		index = location
		row,col = index // nx, index % nx

	
	if(BCs is None):
		BCs = get_normal_BCs(nx,ny)
	gridcpp = dag.GridCPP_f32(nx,ny,dx,dx,3)

	if(MFD):
		mask = np.zeros_like(grid.Z2D,dtype = np.uint8)
		dag.mask_upstream_MFD_f32(gridcpp, mask, tgrid, BCs, row, col)

	else:
		if(stg is None):
			stg = SFGraph(tgrid, BCs = BCs, D4 = True, dx = 1.)
			
		mask = bch.mask_watershed_SFD(index, stg.Stack, stg.Sreceivers).reshape((ny,nx))


	return mask if extra_mask is None else combine_masks(mask,extra_mask)

def _legacy_remove_seas(grid, sea_level = 0., extra_mask = None):

	# Preprocessing the boundary conditions
	gridcpp = dag.GridCPP_f32(grid.nx,grid.ny,grid.dx,grid.dx,3)

	BCs = np.ones_like(grid.Z2D, dtype = np.uint8)
	
	mask = mask_seas(grid, sea_level, extra_mask)

	# nanmask = 
	
	if (extra_mask is None):
		mask = combine_masks(mask, extra_mask)

	return mask_to_BCs(grid,mask)

def remove_seas(grid, sea_level = 0., extra_mask = None):

	if( isinstance(grid, scb.raster.RegularRasterGrid) == False):
		return _legacy_remove_seas(grid, sea_level, extra_mask)

	else:

		# Preprocessing the boundary conditions
		gridcpp = dag.GridCPP_f32(grid.geo.nx, grid.geo.ny, grid.geo.dx, grid.geo.dx, 3)

		BCs = np.ones_like(grid.Z, dtype = np.uint8)
		
		mask = mask_seas(grid, sea_level, extra_mask)
		
		if (extra_mask is not None):
			mask = combine_masks(mask, extra_mask)


		return mask_to_BCs(grid, mask)



def mask_main_basin(grid, sea_level = None, BCs = None, extra_mask = None, MFD = True, stg = None):

	if(sea_level is not None):
		BCs = remove_seas(grid, sea_level, extra_mask = None if (BCs is None) else (BCs > 0))

	if(stg is None):
		stg = scb.flow.SFGraph(grid.Z, BCs = BCs, D4 = False, dx = 1., backend = 'ttb', fill_LM = True, step_fill = 1e-3)

	A = scb.flow.drainage_area(stg).ravel()
	index = np.argmax(A)
	row,col = index // grid.geo.nx, index % grid.geo.nx

	if(BCs is None):
		BCs = scb.flow.get_normal_BCs(grid)

	if(MFD):
		gridcpp = dag.GridCPP_f32(grid.geo.nx,grid.geo.ny,grid.geo.dx,grid.geo.dx,3)
		mask = np.zeros_like(grid.Z,dtype = np.uint8)
		dag.mask_upstream_MFD_f32(gridcpp, mask, grid.Z, BCs, row, col)
	else:
		mask = bch.mask_watershed_SFD(index, stg.Stack, stg.Sreceivers).reshape((grid.geo.ny,grid.geo.nx))
	return mask if extra_mask is None else combine_masks(mask,extra_mask)






'''
WIP for graphflood universale interface
'''
import numpy as np
import scabbard as scb
import scipy.io as sio


class GFUI:

	def __init__(self, backend, grid):


		backend = backend.lower()
		# setting up the backend
		if(backend in ['gpu' , 'dagger' , 'cpu' , 'ttb' , 'topotoolbox']):
			self.backend = backend
		else:
			raise RuntimeError('Backend ' + backend + ' not recognised.')

		#########################
		# internal backend object
		#
		
		## graphflood OG from DAGGER (v1)
		### obj
		self._cpp_gf = None
		### graph
		self._cpp_graph = None
		### graph
		self._cpp_connector = None
		
		## Induced graph and v2
		self._cpp_env = None

		## lib-py-TopoToolBox
		# No object

		## GPU backend
		self._param = None
		self._rd = None


	def _init_riverdale_backend(self, grid):
		pass



		


'''
Sets of function to compute local minimas
EXPERIMENTAL, no warranty it evens do what it is supposed to do yet

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_helper_surfw as srf
import dagger as dag
from scipy.ndimage import gaussian_filter



# (is cpu)
def priority_flood(rd, Zw = True, step = 1e-3):
	'''
	Applies priority flood's Algorithm in D4 topology to the riverdale's elevation.
	Note that it takes into account the boundary conditions.

	Arguments:
		- rd: An initialised RiverDale's instance
		- Zw: If True, fills the water surface, if False, the topography
	Returns:
		- Nothing, edits hte model in place and retransfer the data to GPU
	Authors:
		- B.G. (last modification: 30/05/2024)
	'''

	tZw = rd.Z.to_numpy() + rd.hw.to_numpy() if Zw else rd.Z.to_numpy()
	gcpp = rd.get_GridCPP()
	dag._PriorityFlood_D4_f32(tZw,gcpp,rd.BCs.to_numpy(),step) if rd.param.dtype_float == ti.f32 else dag._PriorityFlood_D4_f64(tZw,gcpp,rd.BCs.to_numpy(),step)
	if(Zw):
		rd.hw.from_numpy(tZw - rd.Z.to_numpy())
	else:
		rd.Z.from_numpy(tZw)

def smooth_hw(rd, strength = 3, recompute_QwA = True):
	'''
	Applies priority flood's Algorithm in D4 topology to the riverdale's elevation.
	Note that it takes into account the boundary conditions.

	Arguments:
		- rd: An initialised RiverDale's instance
		- Zw: If True, fills the water surface, if False, the topography
	Returns:
		- Nothing, edits hte model in place and retransfer the data to GPU
	Authors:
		- B.G. (last modification: 30/05/2024)
	'''
	import scabbard.riverdale.rd_hydrodynamics as rdhy

	tZw = rd.Z.to_numpy() + rd.hw.to_numpy()
	mask = rd.hw.to_numpy() <= 0
	tZw = gaussian_filter(tZw, strength)
	thw = tZw - rd.Z.to_numpy()
	thw[mask] = 0.
	rd.hw.from_numpy(thw)
	if (recompute_QwA):
		rdhy._compute_QwA_from_Zw(rd.Z, rd.hw, rd.QwA, rd.BCs )



@ti.kernel
def N_conv(hw:ti.template(), QwA:ti.template(), QwC:ti.template(), BCs:ti.template(), threshold_QwA:ti.f32, threshold_hw:ti.f32, threshold_conv:ti.f32) -> (ti.i32, ti.i32):

	N:ti.i32 = 0
	NC:ti.i32 = 0

	for i,j in hw:

		if gridfuncs.is_active(i,j,BCs) == False:
			continue

		if QwA[i,j] < threshold_QwA or threshold_hw > hw[i,j]:
			continue

		N += 1

		if(abs(QwC[i,j]/QwA[i,j] - 1) < threshold_conv):
			NC += 1

	return NC,N

def compute_convergence(rd, threshold_QwA = 1e-4, threshold_hw = 1e-2, threshold_conv = 0.05, min_N = 100):
	NC,N = N_conv(rd.hw, rd.QwA, rd.QwC, rd.BCs, threshold_QwA, threshold_hw, threshold_conv)
	print(NC,N)
	if N < min_N:
		return 0
	else:
		return NC/N


@ti.kernel
def constrain_drape(Z:ti.template(), hw:ti.template(), constrains:ti.template(), BCs:ti.template()):
	'''
	TODO to deprecate probably
	'''

	# So, let's try to constrain the min/max height to add without creating local minimas

	for i,j in Z:

		

		tZw = srf.Zw_drape(Z,hw,i,j)

		constrains[i,j,0] = tZw
		constrains[i,j,1] = tZw

		if gridfuncs.is_active(i,j,BCs) == False:
			continue

		# Traversing Neighbours
		for k in range(4):

			# Getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			if(ir == -1):
				continue

			constrains[i,j,0] = ti.math.min(constrains[i,j,0], srf.Zw_drape(Z,hw,ir,jr))
			constrains[i,j,1] = ti.math.max(constrains[i,j,1], srf.Zw_drape(Z,hw,ir,jr))

		constrains[i,j,0] -= Z[i,j]
		constrains[i,j,1] -= Z[i,j]









@ti.kernel
def label_pit(Z:ti.template(), LM:ti.template(), BCs:ti.template()):

	for i,j in Z:

		if(gridfuncs.is_active(i,j,BCs) == False):
			LM[i,j] = 0
			continue

		isLM = True
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue
			if(Z[i,j] > Z[ir,jr]):
				isLM = False
				break
		lab = i * GRID.nx + j
		if(isLM):
			LM[i,j] = lab 


@ti.kernel
def count_pits(Z:ti.template(), BCs:ti.template())  -> ti.i32:

	count = 0

	for i,j in Z:

		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		isLM = True
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue
			if(Z[i,j] > Z[ir,jr]):
				isLM = False
				break

		if(isLM):
			count += 1
	return count

@ti.func
def Zw(Z,hw,i,j):
	return Z[i,j] + hw[i,j]

@ti.kernel
def count_pits_Zw(Z:ti.template(), hw:ti.template(), BCs:ti.template())  -> ti.i32:

	count = 0

	for i,j in Z:

		if(gridfuncs.is_active(i,j,BCs) == False or gridfuncs.can_out(i,j,BCs)):
			continue

		isLM = True
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# if(srf.Zw_drape(Z,hw,i,j) == srf.Zw_drape(Z,hw,ir,jr)):
			# 	print('happens')

			if( srf.Zw_drape(Z,hw,i,j) > srf.Zw_drape(Z,hw,ir,jr) ) :
				isLM = False
				break

		if(isLM):
			count += 1

	return count


@ti.kernel
def label_pits_Zw(Z:ti.template(), hw:ti.template(), LM:ti.template(), BCs:ti.template())  -> ti.i32:

	count = 0

	for i,j in Z:
		LM[i,j] = 0

		if(gridfuncs.is_active(i,j,BCs) == False or gridfuncs.can_out(i,j,BCs)):
			continue

		isLM = True
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# if(srf.Zw_drape(Z,hw,i,j) == srf.Zw_drape(Z,hw,ir,jr)):
			# 	print('happens')

			if( srf.Zw_drape(Z,hw,i,j) > srf.Zw_drape(Z,hw,ir,jr) ) :
				isLM = False
				break

		if(isLM):
			LM[i,j] = 1
			count += 1

	return count




#############################################################################################################
#############################################################################################################
#############################################################################################################
#############################################################################################################



# Bellow are experimental stuff


#############################################################################################################
#############################################################################################################
#############################################################################################################
#############################################################################################################







@ti.kernel
def label_depression(LM:ti.template(), D4dir:ti.template(), BCs:ti.template(), count:ti.template()):

	count[None] = 0

	for i,j in LM:
		
		if(LM[i,j] == 0):
			continue

		ti = i
		tj = j

		lab = LM[i,j]

		while(True):
			tti = ti
			ttj = tj

			for k in range(4):
				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(ti, tj, k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1 or LM[ir,jr]):
					continue
				
				ik = 3 - k
				if(D4dir[ir,jr] != ik):
					continue

				LM[ir,jr] = lab
				ti,tj = ir,jr
				count[None] += 1
				# break

			if(ti == tti and tj == ttj):
				break
@ti.kernel
def compute_depressions_min_pass(Z:ti.template(), LM:ti.template(), Z_LM_min:ti.template(), BCs:ti.template()):
	'''
	'''

	for i,j in Z:
		if(LM[i,j] == 0):
			continue

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue
			if(LM[ir,jr] == LM[i,j]):
				continue

			ti.atomic_min(Z_LM_min[LM[i,j]], max(Z[i,j],Z[ir,jr]) + 1e-3);

@ti.kernel
def raise_depressions(Z:ti.template(), LM:ti.template(), Z_LM_min:ti.template(), BCs:ti.template()):
	'''
	'''

	for i,j in Z:
		if(LM[i,j] == 0):
			continue

		Z[i,j] = max(Z_LM_min[LM[i,j]], Z[i,j])

@ti.kernel
def detect_flats(Z:ti.template(), D4dir:ti.template(), isflat:ti.template(), BCs:ti.template()):

	for i,j in Z:
		
		allflat = True
		
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			if(Z[ir,jr] != Z[i,j]):
				allflat = False
				break

		isflat[i,j] = allflat


	for i,j in Z:
		
		if(gridfuncs.is_active(i,j,BCs) == False or isflat[i,j] == False):
			continue

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			if(Z[ir,jr] == Z[i,j]):
				isflat[ir,jr] = True

	for i,j in Z:
		if(isflat[i,j]):
			D4dir[i,j] = 5

@ti.kernel
def label_flats(Z:ti.template(), LM:ti.template(), isflat:ti.template(), BCs:ti.template(), count:ti.template()):

	label = 0
	for i,j in Z:

		if(is_flat[i,j] == False):
			continue


		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue











@ti.kernel
def resolve_flats(Z:ti.template(), D4dir:ti.template(), isflat:ti.template(), BCs:ti.template(), count:ti.template()):
	'''
		Does not work yet
	'''

	count[None] = 0
	for i,j in Z:
		if(isflat[i,j] == False):
			continue

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)
			if(ir == -1):
				continue

			if(Z[ir,jr] < Z[i,j] and D4dir[i,j] == 5):
				D4dir[i,j] = k
				count[None] += 1

			# elif(D4dir[ir,jr] == 5 and D4dir[i,j] != 5):
			# 	D4dir[ir,jr] = 3 - k
			# 	count[None] += 1
			
			elif(D4dir[ir,jr] != 5 and D4dir[i,j] == 5 and isflat[ir,jr] and D4dir[ir,jr] != 3 - k):
				D4dir[i,j] = k
				count[None] += 1





@ti.kernel
def label_oedges(edges:ti.template(), BCs:ti.template()):

	for i,j in edges:
		if(gridfuncs.can_out(i,j,BCs)):
			edges[i,j] = 1
		else:
			edges[i,j] = 0


@ti.kernel
def label_edges(edges:ti.template(), D4dir:ti.template(), BCs:ti.template(), count:ti.template()):

	count[None] = 0

	for i,j in edges:
		
		if(edges[i,j] == 0):
			continue

		ti = i
		tj = j

		while(True):
			tti = ti
			ttj = tj

			for k in range(4):
				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(ti, tj, k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1 or edges[ir,jr]):
					continue
				
				ik = 3 - k
				if(D4dir[ir,jr] != ik):
					continue

				edges[ir,jr] = 1
				ti,tj = ir,jr
				count[None] += 1
				# break

			if(ti == tti and tj == ttj):
				break

@ti.kernel
def reconstruct(edges:ti.template(), D4dir:ti.template(), Z:ti.template(), BCs:ti.template(), count:ti.template() ) :
	'''
	DOES NOT WORK
	'''

	count[None] = 0

	for i,j in edges:
		
		if(edges[i,j] > 0):
			continue

		tZ = Z[i,j]
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i, j, k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1 or edges[ir,jr] == 0):
				continue

			if(Z[ir,jr] > Z[i,j] and Z[ir,jr] <= tZ):
				tZ = Z[ir,jr] 
		
		if(tZ != Z[i,j]):
			count[None] += 1
			edges[i,j] = 1
			Z[i,j] = tZ + 1e-3
			
			# break


	

@ti.kernel
def archive_1_label_edges(edges:ti.template(), D4dir:ti.template(), BCs:ti.template(), count:ti.template()):
	'''
	this experimentation on label edges propagates from neighbours to edges
	'''
	count[None] = 0
	# inei = ti.math.ivec4(3,2,1,0)

	for i,j in edges:
		
		if(edges[i,j] == 0):
			continue

		ti = i
		tj = j

		# while(True):

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(ti, tj, k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1 or edges[ir,jr]):
				continue
			
			ik = 3 - k
			if(D4dir[ir,jr] != ik):
				continue

			edges[ir,jr] = 1
			count[None] += 1

	




@ti.kernel
def forward_scan( marker : ti.template(), mask:ti.template(), BCs:ti.template(), count : ti.template() ):
	

	for i,j in marker:

		if(gridfuncs.is_active(i,j,BCs) == False):
			continue
		# Compute the maximum of the marker at the current pixel and all
		# of its previously visisted neighbors
		max_height = marker[i,j]
		# Traversing Neighbours
		for k in range(2):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			if max_height > marker[ir,jr]:
				max_height = max_height  
			else:
				max_height = marker[ir,jr]

		# Set the marker at the current pixel to the minimum of the
		# maximum height of the neighborhood and the mask at the current
		# pixel.
		z = max_height  

		if max_height >= mask[i,j]:
			# print(i,j)
			z = mask[i,j]

		if (z != marker[i,j]):
			print(i,j)

			# Increment count only if we change the current pixel
			count[None] += 1; # Should be atomic
			marker[i,j] = z;



@ti.kernel
def backward_scan( marker : ti.template(), mask:ti.template(), BCs:ti.template(), count : ti.template() ):
	# Note that the loop decreases. p must have a signed type for this
	# to work correctly.
	for i,j in marker:

		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Compute the maximum of the marker at the current pixel and all
		# of its previously visisted neighbors
		max_height = marker[i,j]
		# Traversing Neighbours
		for tk in range(2):
			k = tk + 1
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue
			max_height = max_height if max_height > marker[ir,jr] else marker[ir,jr]
	

	# Set the marker at the current pixel to the minimum of the
	# maximum height of the neighborhood and the mask at the current
	# pixel.

		z = max_height if max_height < mask[i,j] else mask[i,j]
		if (z != marker[i,j]):
			# Increment count only if we change the current pixel
			count[None] += 1;
			marker[i,j] = z;




@ti.kernel
def prefillsink(Z:ti.template(), Z_fill:ti.template(), BCs:ti.template()):
	'''
	Mimics the fillsinks function of libtopotoolbox
	Does not work yet
	'''

	for i,j in Z:
		Z[i,j] *= -1 
		if(gridfuncs.is_active(i,j,BCs)):
			Z_fill[i,j] = Z[i,j]

@ti.kernel
def postfillsink(Z:ti.template(), Z_fill:ti.template(), BCs:ti.template()):
	'''
	Mimics the fillsinks function of libtopotoolbox
	Does not work yet
	'''

	for i,j in Z:
		Z[i,j] *= -1 
		Z_fill[i,j] *= -1


def fillsinks(Z,Z_fill,BCs):

	count = ti.field(dtype = ti.f32, shape = ())

	prefillsink(Z,Z_fill,BCs)

	count[None] = 1

	while(count[None] != 0):
		print("run", count[None])
		count[None] = 0
		forward_scan( Z_fill, Z, BCs, count)
		backward_scan( Z_fill, Z, BCs, count)
		print("run", count[None])

	postfillsink(Z,Z_fill,BCs)




# Early tests

@ti.func
def Z_draped(i:ti.int32, j:ti.int32, Z:ti.template(), hw:ti.template()):
	return Z[i,j] + hw[i,j]


@ti.kernel
def pre_drape(Z:ti.template(), hw:ti.template(), recConstrain: ti.template(), donorConstrain:ti.template(), BCs:ti.template(), count:ti.template()):
	'''
		stuff
	'''
	#
	count[None] = 0

	for i,j in Z:
		recConstrain[i,j] = 1e9
		donorConstrain[i,j] = -1e9

	for i,j in Z:

		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		checker = True

		for k in range(4):

			# Getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# If not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# computing the minimum height of the donors in order not to invert 
			# Note that I am trying to get the minimum elevation of the donor
			if (Z_draped(ir,jr,Z,hw) > Z_draped(i,j,Z,hw) and (Z_draped(ir,jr,Z,hw) < donorConstrain[i,j] or donorConstrain[i,j] == -1e9)):
				donorConstrain[i,j] = Z_draped(ir,jr,Z,hw)
				checker = False

			# Computing the maximum drop in elevation, in this case this is the lowest receiver
			elif (Z_draped(ir,jr,Z,hw) < Z_draped(i,j,Z,hw) and (Z_draped(ir,jr,Z,hw) < recConstrain[i,j] or recConstrain[i,j] == 1e9)):
				recConstrain[i,j] = Z_draped(ir,jr,Z,hw)
				checker = False
		
		if(checker):
			count[None] += 1


	# Done





@ti.kernel
def drape_iteration(Z:ti.template(), hw:ti.template(), recConstrain: ti.template(), donorConstrain:ti.template(), BCs:ti.template(), count:ti.template()):
	'''
	stuff 2
	'''
	count[None] = 0
	for i,j in Z:

		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		temp1 = Z_draped(i,j,Z,hw)
		temp1 = ti.min(ti.max(temp1, recConstrain[i,j]), donorConstrain[i,j])
		# ti.math.clamp(temp1, recConstrain[i,j], donorConstrain[i,j])
		temp2 = Z_draped(i,j,Z,hw)
		if(temp1 != temp2):
			count[None] += 1












































































# end of file 


'''
Contains routines to create raster grids from scratch.

For example white noise, perlin noise or spectral stuff for initial conditions,
or constant slope raster to test river LEMs, ...

B.G. - created 09/2024
'''

import scabbard as scb
import numpy as np
import matplotlib.pyplot as plt
import taichi as ti
import numba as nb
import time


def slope2D_S(
	nx = 256, 
	ny = 512, 
	dx = 2.,
	z_base = 0,
	slope = 1e-3, 
	):
	'''
		Generates a slopping grid adn initialise an env with given boundaries to it. 
		It comes with a connector and a graph (todo) alredy pre-prepared for boundaries
	'''

	# Zeros Topo
	Z = np.zeros((ny, nx), dtype = np.float32)

	lx = (nx+1) * dx
	ly = (ny+1) * dx

	grid = scb.raster.raster_from_array(Z, dx=dx, xmin=0.0, ymin=0.0, dtype=np.float32)

	XX,YY = grid.geo.XY
	
	Z = (YY * slope)[::-1]
	grid.Z[:,:] = Z[:,:]

	BCs = np.ones((ny,nx),dtype = np.uint8)
	BCs[:,[0,-1]] = 0
	BCs[0,:] = 7
	BCs[-1,:] = 3

	return grid, BCs

def white_noise(nx, ny, dx, magnitude = 1., BCs = None):

	Z = np.random.rand(ny,nx) * magnitude
	grid = scb.raster.raster_from_array(Z, dx=dx, xmin=0.0, ymin=0.0, dtype=np.float32)

	if(BCs is None):
		BCs = scb.ut.normal_BCs_from_shape(nx,ny)
	elif isinstance(BCs, str):
		if(BCs == '4edges'):
			BCs = scb.ut.normal_BCs_from_shape(nx,ny)
		elif(BCs == 'periodicNS'):
			BCs = scb.ut.periodic_NS_BCs_from_shape(nx,ny)
		elif(BCs == 'periodicEW'):
			BCs = scb.ut.periodic_NS_BCs_from_shape(nx,ny)
		else:
			raise ValueError('scabbard.raster.raster_factory.white_noise::BCs is string but not 4edges, periodicNS or periodicEW')
	elif isinstance(BCs,scb.raster.RegularRasterGrid):
		grid.Z[BCs == 0] = 0.
	else:
		raise ValueError('scabbard.raster.raster_factory.white_noise::BCs is not recognised')

	return grid, BCs


'''
This script contains set of routines to manage local minima using different methods and/or libraries


B.G
'''

import numpy as np
import scabbard as scb
import scabbard._utils as ut
import scabbard.flow.graph as gr
import dagger as dag
import scabbard.flow._LM as lmn
import topotoolbox as ttb


def _priority_flood_from_Z(Z, BCs, D4, in_place, dx, gridcpp, backend, step_fill):

	if(in_place):
		tZ = Z
	else:
		tZ = Z.copy()

	if(BCs is None):
		BCs = ut.normal_BCs_from_shape(Z.shape[1], Z.shape[0])

	if(backend == 'ttb'):
		dims = np.array( [Z.shape[0], Z.shape[1]],dtype = np.uint64)
		# ttb.graphflood.funcdict['priority_flood'](Z.ravel(), BCs.ravel(), dims, not D4)
		ttb.graphflood.funcdict['priority_flood_TO'](Z.ravel(), np.zeros_like(Z, dtype = np.uint64), BCs.ravel(), dims, not D4, step_fill)

	elif backend == 'dagger':
		if(gridcpp is None):
			gridcpp = dag.GridCPP_f32(Z.shape[1],Z.shape[0], dx, dx,3)
		dag._PriorityFlood_D4_f32(tZ, gridcpp, BCs, 1e-4)

	if(in_place == False):
		return tZ

def _priority_flood_from_dem(dem, BCs, D4, in_place, dx, gridcpp, backend, step_fill):

	if(in_place):
		tZ = dem.Z
	else:
		# print("COPY")
		tZ = dem.Z.copy()

	if(BCs is None):
		BCs = ut.normal_BCs_from_shape(tZ.shape[0], tZ.shape[1])

	if(backend == 'ttb'):
		dims = np.array( [tZ.shape[0], tZ.shape[1]],dtype = np.uint64)
		ttb.graphflood.funcdict['priority_flood'](tZ.ravel(), BCs.ravel(), dem.dims, not D4, step_fill)
		# ttb.graphflood.funcdict['priority_flood_TO'](tZ.ravel(), np.zeros_like(tZ, dtype = np.uint64), BCs.ravel(), dims, not D4, step_fill)

	elif backend == 'dagger':
		if(gridcpp is None):
			gridcpp = dag.GridCPP_f32(tZ.shape[1],tZ.shape[0], dx, dx,3)
		dag._PriorityFlood_D4_f32(tZ, gridcpp, BCs, 1e-4)

	if(in_place == False):
		return dem.duplicate_with_other_data(tZ)
	else:
		dem.Z[:,:] = tZ[:,:]

def priority_flood(Z, BCs = None, D4 = True, in_place = True, dx = 1., gridcpp = None, backend = 'ttb', step_fill = 1e-3):
	'''
	perform priority flood + slope on a 2D DEM
	'''

	if(isinstance(Z,np.ndarray)):
		return _priority_flood_from_Z(Z, BCs, D4, in_place, dx, gridcpp, backend, step_fill)
	elif(isinstance(Z,scb.raster.RegularRasterGrid)):
		return _priority_flood_from_dem(Z, BCs, D4, in_place, dx, gridcpp, backend, step_fill)

	

def legacy_break_bridges(grid, in_place = False, BCs = None, step_fill = 1e-3):
	'''
	Experimental function to break bridges and local minimas in a general way
	
	argument:
		- grid: A Rgrid object (TODO adapt to new grid systems)

	B.G.
	'''

	Z = grid.Z2D.copy()

	if(BCs is None):
		BCs = ut.normal_BCs_from_shape(grid.nx,grid.ny)

	gridcpp = dag.GridCPP_f32(Z.shape[1],Z.shape[0], grid.dx, grid.dx, 3)

	# first filling the topo
	filled_Z = priority_flood(Z, BCs = BCs, in_place = False, gridcpp = gridcpp, dx = grid.dx, step_fill = step_fill)

	# COmputing a first graph
	sgf = gr.SFGraph(filled_Z, BCs = None, D4 = True, dx = grid.dx)

	lmn.impose_downstream_minimum_elevation_decrease(Z.ravel(), sgf.Stack, sgf.Sreceivers.ravel(), delta = step_fill)

	return Z
	


def break_bridges(grid, in_place = False, BCs = None, step_fill = 1e-3):

	if(isinstance(grid,scb.raster.RegularRasterGrid) == False):
		return legacy_break_bridges(grid, in_place = False, BCs = None, step_fill = 1e-3)


	Z = grid.Z.copy() if not in_place else grid.Z

	if(BCs is None):
		BCs = ut.normal_BCs_from_shape(grid.nx,grid.ny)


	# first filling the topo
	filled_Z = priority_flood(Z, BCs = BCs, in_place = False, dx = grid.geo.dx, step_fill = step_fill)

	# COmputing a first graph
	sgf = gr.SFGraph(filled_Z, BCs = None, D4 = True, dx = grid.geo.dx)

	lmn.impose_downstream_minimum_elevation_decrease(Z.ravel(), sgf.Stack, sgf.Sreceivers.ravel(), delta = step_fill)

	if(in_place == False):
		return Z











'''
Contains generic widgets customisation for matplotlib qt backend
For example specific types of sliders (float, min max)
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

# from PySide6 import QtWidgets, QtCore
from matplotlib.backends.qt_compat import QtWidgets, QtCore
import matplotlib
# Ensure using the Qt6Agg backend with PySide6
matplotlib.use('QtAgg')








'''
this script defines the geometry features of given elements
They are made to be accessed from the top level (end users)


B.G.
'''

import numpy as np
from .geometry import BaseGeometry

class RegularGeometry(BaseGeometry):
	"""
	geometry class for any object with a regular 2D structure
	that can be described with a number of rows and col

	Authors:
		- B.G. (last modifications: 08/2024)


	"""

	def __init__(self, nx, ny, dx, xmin, ymin):
		'''
		Creates the Regular Geometry object

		Arguments:
			- nx: number of columns
			- ny: number of rows
			- dx: spatial step
			- xmin: minimum Easting coordinate
			- ymin: minimum Northing coordinate
		
		Returns:
			- an instance of RegularGeometry

		Authors:
			- B.G. (last modification: 08/2024)
		'''

		super().__init__()

		self._nx = nx
		self._ny = ny
		self._nxy = nx * ny
		self._dx = dx
		self._lx = dx * (nx+1)
		self._ly = dx * (ny+1)
		self._dxy = 2**0.5 * dx
		self._xmin = xmin
		self._ymin = ymin


	@property
	def dx(self):
		'''
		return the spatial step

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._dx

	@property
	def dxy(self):
		'''
		return the spatial step

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._dxy

	@property
	def N(self):
		'''
		return the number of nodes in the element

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._nxy


	@property
	def nx(self):
		'''
		return the number of nodes in the x directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._nx


	@property
	def ny(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._ny

	@property
	def lx(self):
		'''
		return the length in the x directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._lx


	@property
	def ly(self):
		'''
		return the length in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._ly

	def shape(self):
		'''
		Returns a numpy=like shape
		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return (self.ny,self.nx)


	@property
	def xmin(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._xmin

	@property
	def xmax(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._xmin + (self.nx + 1) * self.dx

	@property
	def ymin(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._ymin

	@property
	def ymax(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self._ymin + (self.ny + 1) * self.dx



	def row_col_to_flatID(self, row, col):
		'''
		Take row col (single or array) and returns the flat index for regular datas

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return row * self.nx + col

	def flatID_to_row_col(self, flatID):
		'''
		Take row col (single or array) and returns the flat index for regular datas

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return flatID // self.nx, flatID % self.nx


	def row_col_to_X_Y(self, row, col):
		'''
		Converts row col (for regular grids) to X Y coordinates (real world)

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		if(self.Y_inverted):
			return (col * self.dx) + self.xmin + self.dx/2, ( (self.ny - 1 - row) * self.dx) + self.ymin + self.dx/2 
		else:
			return (col * self.dx) + self.xmin + self.dx/2, (row * self.dx) + self.ymin + self.dx/2



	def X_Y_to_row_col(self, X, Y):
		'''
		Converts  X Y coordinates (real world) to row col (for regular grids)

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		if(self.Y_inverted):
			return int((self.ny - 1 - np.floor(Y - self.ymin))/self.dx), int(np.floor(X - self.xmin)/self.dx)
		else:
			return int(np.floor(Y - self.ymin)/self.dx), int(np.floor(X - self.xmin)/self.dx)

	def flatID_to_X_Y(self, flatID):
		'''
		Converts  flat ID to XY coordinates

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		row,col = self.flatID_to_row_col(flatID)
		return self.row_col_to_X_Y(row,col)
		

	def X_Y_to_flatID(self, X, Y):
		'''
		Converts  X Y coordinates (real world) to flat ID (for regular grids)

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		row,col = self.X_Y_to_row_col(X,Y)
		return self.row_col_to_flatID(row,col)

	@property
	def xl_centered(self):
		return self.xmin+self.dx/2.

	@property
	def xr_centered(self):
		return self.xmax-self.dx/2.

	@property
	def yt_centered(self):
		return self.ymin+self.dx/2.

	@property
	def yb_centered(self):
		return self.ymax-self.dx/2.

	@property
	def X(self):
		'''
		array of X indices at the center of nodes
		'''
		return np.linspace(self.xl_centered,self.xr_centered, self.nx)

	@property
	def Y(self):
		'''
		array of Y indices at the center of nodes
		'''
		return np.linspace(self.yt_centered, self.yb_centered, self.ny)

	@property
	def Y_inverted(self):
		'''
		return if Y is inverted (top to down) or not (down to top)
		'''
		tY = self.Y
		if(tY[0] > tY[-1]):
			return True
		else:
			return False
	@property
	def XY(self):
		'''
		return a mesh grid of X and Y coorditates
		i.e. 2 2d arrays of Xs and Ys for each nodes
		'''
		return np.meshgrid(self.X, self.Y)

	@property
	def dxnxny(self):
		return self.dx,self.nx,self.ny	 




'''
Contains generic widgets customisation for matplotlib qt backend
For example specific types of sliders (float, min max)
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

# from PySide6 import QtWidgets, QtCore
from matplotlib.backends.qt_compat import QtWidgets, QtCore
import matplotlib
# Ensure using the Qt6Agg backend with PySide6
matplotlib.use('QtAgg')



class FloatSlider(QtWidgets.QSlider):
	'''
	A widget that emulate a floating point slider
	'''

	valMod = QtCore.Signal((float))

	def __init__(self, orientation, step = 0.5, multiplier = 100, parent=None, vmin = 0, vmax = 42, label = ""):
		super(FloatSlider, self).__init__(orientation, parent)
		self.layout = QtWidgets.QVBoxLayout(self)
		self.layout.addWidget(QtWidgets.QLabel(label))

		self._multiplier = multiplier  # Define your own multiplier for float precision
		self._step = step  # Define the step value you want for the slider
		self.setFloatRange(vmin,vmax)
		self.valueChanged.connect(lambda x: self.valMod.emit(self.floatValue()))



	def setFloatRange(self, min_val, max_val):
		self.setMinimum(int(min_val * self._multiplier))
		self.setMaximum(int(max_val * self._multiplier))

	def setFloatValue(self, value):
		intValue = int(value * self._multiplier)
		self.setValue(intValue)

	def floatValue(self):
		return self.value() / self._multiplier

	def setFloatStep(self, step):
		self._step = step
		self.setSingleStep(int(step * self._multiplier))


class RangeSlider(QtWidgets.QWidget):

	rangeChanged = QtCore.Signal((float, float, str))

	def __init__(self, minimum, maximum, parent=None, ID = "", orientation = QtCore.Qt.Horizontal):
		super(RangeSlider, self).__init__(parent)
		self.layout = QtWidgets.QVBoxLayout(self)

		self.minimum = minimum
		self.maximum = maximum



		self.ID = str(ID)

		self.minSlider = FloatSlider(orientation, self)
		self.maxSlider = FloatSlider(orientation, self)



		self.minSlider.setFloatRange(minimum,maximum)
		self.maxSlider.setFloatRange(minimum,maximum)

		self.minSlider.setFloatValue(minimum)
		self.maxSlider.setFloatValue(maximum)

		# self.minSlider.setMinimum(minimum)
		# self.minSlider.setMaximum(maximum)
		# self.maxSlider.setMinimum(minimum)
		# self.maxSlider.setMaximum(maximum)

		self.layout.addWidget(QtWidgets.QLabel("Min:"))
		self.layout.addWidget(self.minSlider)
		self.layout.addWidget(QtWidgets.QLabel("Max:"))
		self.layout.addWidget(self.maxSlider)

		self.minValue = minimum
		self.maxValue = maximum

		self.minSlider.valueChanged.connect(self.updateRange)
		self.maxSlider.valueChanged.connect(self.updateRange)

		

	def updateRange(self):
		minVal = self.minSlider.floatValue()
		maxVal = self.maxSlider.floatValue()

		if minVal > maxVal:
			minVal, maxVal = maxVal, minVal

		self.minValue = minVal
		self.maxValue = maxVal

		# Update the plot or imshow limits here
		# Make sure the sliders stay within each other's bounds
		# self.minSlider.setMaximum(self.maxValue)
		# self.maxSlider.setMinimum(self.minValue)
		self.minSlider.setFloatValue(minVal)
		self.maxSlider.setFloatValue(maxVal)

		self.rangeChanged.emit(self.minValue, self.maxValue, self.ID)


from .rd_env import *
from .rd_params import *
from .rd_hillshading import hillshading, std_hillshading
from .rd_LM import priority_flood, smooth_hw, N_conv, compute_convergence
from .rd_hydrometrics import *
from . import rd_helper_surfw as helper
